/*******************************************************************************
*
* XNF driver from a lofig.
* Based on Nathtalie Dictus and Eudes Prado Lopez version for FPGA
* By Frederic Petrot
*
*******************************************************************************/
#include <stdio.h>
#include <string.h>
#include MUT_H
#include MLO_H

static char *spec_ibuf[] = { "inff", "inlat", "bufgp", "bufgs", "bufg" };
static int spec_ibuf_size = sizeof spec_ibuf / sizeof ( char * ) ;

static char *clkbuf = "bufg";
static char *ibuf = "ibuf";
static char *obuf = "obuf";
static char *tbuf = "tbuf";
static char *obuft = "obuft";

static char buffer1[BUFSIZE]; /* large buffer for string manipulation */
static char buffer2[BUFSIZE]; /* large buffer for string manipulation */

static int prim_map ; /* mode to map pins o->O, i0->0, and add INVS */
static char prim_input_size ;
static char prim_bubble_count ;

/* forward slash denotes hier for sig and symbol names in a
   flattened XNF file
   REF Figure 2-5 of April 1994 XACT reference guide, Volume II
   XNFMerge section page 2-16
   "XNFMERGE places the instance name of the symbol referencing the
   logic and a / at the beginning of each signal and symbol name"
   Thus .'s are replaced with /'s via dot_to_slash
*/

static char *dot_to_slash(name,buffer)
char *name, *buffer;
{
   int i;
   for ( i=0 ; ; i++ ) {
      if( *(name+i) == '.' ) {
          *(buffer+i) =  '/' ;
      } else {
          *(buffer+i) =  *(name+i) ;
      }
      if ( *(name+i) == NULL ) break ;
   }
   return buffer ;
}

static char *xnfvector( name, buffer )
char *name, *buffer;
{
char *s;
int l ;
   dot_to_slash(name,buffer);  /* buffer == name with dots mapped to slash */
   l = strlen(buffer) ;
   if ((s = strchr(buffer, ' ')) == NULL)
      return buffer;

   buffer[s - buffer] = '<';
   buffer[l++] = '>';
   buffer[l] = '\0';
   return buffer;
}   

int prim_add_inv(name)
char *name;
{
  char diff,pin;
  if ( (! prim_map) || (prim_bubble_count==0) ) return 0;
  diff = prim_input_size - prim_bubble_count ;
  if      ( ! strcmp(name,"o")  ) { return 0 ; }  
  else if ( ! strcmp(name,"i0") ) { pin=0 ; }
  else if ( ! strcmp(name,"i1") ) { pin=1 ; }
  else if ( ! strcmp(name,"i2") ) { pin=2 ; }
  else if ( ! strcmp(name,"i3") ) { pin=3 ; }
  else if ( ! strcmp(name,"i4") ) { pin=4 ; }
  else  { fprintf(stderr,"FORMAT_ERROR: pin name *%s* not in set o,i0,i1,i2,i3,i4 \n",name) ; }
  if ( pin >= diff ) { return 1 ; }
  else { return 0 ; } 
}

static char *prim_map_pin_name(name, buffer)
char *name, *buffer;
{
  xnfvector( name, buffer ) ; /* buffer updated "name elt" to "name<elt>" */
  if ( ! prim_map ) return buffer ;
  if      ( ! strcmp(buffer,"o")  ) { return "O" ; }
  else if ( ! strcmp(buffer,"i0") ) { return "1" ; }
  else if ( ! strcmp(buffer,"i1") ) { return "2" ; }
  else if ( ! strcmp(buffer,"i2") ) { return "3" ; }
  else if ( ! strcmp(buffer,"i3") ) { return "4" ; }
  else if ( ! strcmp(buffer,"i4") ) { return "5" ; }
  else { fprintf(stderr,"FORMAT_ERROR: pin name *%s* not in set o,i0,i1,i2,i3,i4 \n",buffer) ; }
}

static char *xnf_map_figname(name, buffer)
char *name, *buffer;
{
   int l = strlen(name);
   int stpos=3;
   char *sp;
   prim_map=0;
   prim_bubble_count=0;
   /* fd and dff primitives+ to dff, no prim_map */
   if ( ! ( strcmp(name,"fdce") && strcmp(name,"fdpe") && strcmp(name,"fd") ) ) {
        strcpy(buffer,"dff");
        return buffer ;
   }
   /* process plain primitives */
   else if ( strncmp(name,"and",3) == 0 ) {
        strcpy(buffer,"and");
   } else if ( strncmp(name,"nand",4) == 0 ) {
        strcpy(buffer,"nand");
        stpos=4;
   } else if ( strncmp(name,"xor",3) == 0 ) {
        strcpy(buffer,"xor");
   } else {
      return name;
   }
   sp = name + stpos ;
   if ( l == stpos + 1 ) {
      prim_input_size =  *sp - 0x30 ;
      if ( (prim_input_size < 2) || (prim_input_size > 5) ) { fprintf(stderr,"A:FORMAT_ERROR\n") ; }
   } else if ( l == stpos+2+1) {
      prim_input_size =  *sp - 0x30 ;
      if ( (prim_input_size < 2) || (prim_input_size > 5) ) { fprintf(stderr,"B:FORMAT_ERROR\n") ; }
      if ( strncmp(++sp,"b",1) != 0 ) {  fprintf(stderr,"C:FORMAT_ERROR name=%s\n",name) ; }
      prim_bubble_count=*(++sp) - 0x30 ;
      if ( (prim_bubble_count < 0) || (prim_bubble_count > prim_input_size) ) { 
         fprintf(stderr,"D:FORMAT_ERROR\n") ;
      }
   } else { fprintf(stderr,"E:FORMAT_ERROR name=%s l=%d stpos=%d \n",name,l,stpos) ; }

   prim_map = 1;
   return buffer ;
}

static lofig_list *mkbuf()
{
lofig_list *lof = NULL;
locon_list *lc;
losig_list *ls;

   /* let us generate the buffers that are necessary for our IOB */
   lof = addlomodel(lof, ibuf);
   addlocon(lof, "vss",
            addlosig(lof, 0, addchain(NULL, "vss"), EXTERNAL, 0.0), IN);
   addlocon(lof, "vdd",
            addlosig(lof, 1, addchain(NULL, "vdd"), EXTERNAL, 0.0), IN);
   addlocon(lof, "o",
            addlosig(lof, 2, addchain(NULL, "o"), EXTERNAL, 0.0), OUT);
   addlocon(lof, "i",
            addlosig(lof, 3, addchain(NULL, "i"), EXTERNAL, 0.0), IN);
   ibuf = lof->NAME; /* avoid the double namealloc! */

   lof = addlomodel(lof, obuf);
   addlocon(lof, "vss",
            addlosig(lof, 0, addchain(NULL, "vss"), EXTERNAL, 0.0), IN);
   addlocon(lof, "vdd",
            addlosig(lof, 1, addchain(NULL, "vdd"), EXTERNAL, 0.0), IN);
   addlocon(lof, "o",
            addlosig(lof, 2, addchain(NULL, "o"), EXTERNAL, 0.0), OUT);
   addlocon(lof, "i",
            addlosig(lof, 3, addchain(NULL, "i"), EXTERNAL, 0.0), IN);
   obuf = lof->NAME; /* avoid the double namealloc! */

   lof = addlomodel(lof, obuft);
   addlocon(lof, "vss",
            addlosig(lof, 0, addchain(NULL, "vss"), EXTERNAL, 0.0), IN);
   addlocon(lof, "vdd",
            addlosig(lof, 1, addchain(NULL, "vdd"), EXTERNAL, 0.0), IN);
   addlocon(lof, "o",
            addlosig(lof, 2, addchain(NULL, "o"), EXTERNAL, 0.0), OUT);
   addlocon(lof, "t",
            addlosig(lof, 3, addchain(NULL, "t"), EXTERNAL, 0.0), IN);
   addlocon(lof, "i",
            addlosig(lof, 4, addchain(NULL, "i"), EXTERNAL, 0.0), IN);
   obuft = lof->NAME; /* avoid the double namealloc! */
   tbuf = namealloc(tbuf);

   return lof;

}

static char *signalname(h, lof, ls, buffer)
ht *h;
lofig_list *lof;
losig_list *ls;
char *buffer;
{
int  val;
char *s = getsigname(ls);

   if ((val = gethtitem(h, s)) == EMPTYHT) /* the name is unique up to now */
      addhtitem(h, s, (int)ls);
      /* that suck because I assume sizeof(int) == sizeof (void *),
         which insn't true on some beasts */
   else if (val != (int)ls) {
      /* that sucks: two signals have same name, so I return the
         concatenation of the signal name and its index. */
      sprintf(buffer, "%s%d", s, ls->INDEX);
      s = namealloc(buffer);
   }
   return xnfvector(s,buffer);
}

static char dir(lc)
locon_list *lc;
{
   switch(lc->DIRECTION) {
      case IN: return 'I';
      case TRISTATE: return 'B';
      case INOUT:
      case OUT: return 'O';
      default: return 'X';
   }
}

void xnfsavelofig(lof)
lofig_list *lof;
{
FILE *file;
loins_list *li;
locon_list *lc;
losig_list *ls;
long index, hsize = 0;
losig_list *sigvdd = NULL, *sigvss = NULL;
lofig_list *buflist;
ht *h; /* used to check signal name unicity no warrantied by mbk */
int i ;

   if ((file = mbkfopen(lof->NAME, "xnf", "w")) == NULL) {
      fflush(stdout);
      fprintf(stderr, "mbk error : unable to open file %s/%s.%s",
              WORK_LIB, lof->NAME, "xnf");
      EXIT(1);
   }

   VDD = namealloc(VDD);
   VSS = namealloc(VSS);
   buflist = mkbuf();
   lofigchain(lof);

   fprintf(file, "LCANET,4\n");
   fprintf(file, "USER,xnf file %s generated for %s\n",
                  lof->NAME, getenv("USER"));
   fprintf(file, "PROG,mbk2xnf,0.0,\"Mapping for X3000 macros only!\"\n");
   fprintf(file, "PART,3030pc84-125\n");
   fprintf(file, "PWR,0,%s\n", VSS);
   fprintf(file, "PWR,1,%s\n", VDD);

   /* This is a very simple and brute force xnf driver:
      detecting the IO is a pain, since the macro library may be
      enhanced by the user, and that IOB cannot be equivalent to CLB
      from a macro point of view.
      Therefore, all inputs are buffurized, all ouputs are bufferized,
      and tree state outputs are changed to iob for tree state. */
   /* waranty correct external signal detection, who knows... */
   for (lc = lof->LOCON; lc; lc = lc->NEXT) {
   chain_list *c;
   locon_list *externalcon;
   int onexternalbuf;
#     ifdef DEBUG
         /* lc->SIG points into a link list of losigs belonging to all the locons,
            only the one immediately pointed at belongs to the connector at hand */
         (void)fprintf(stdout,"locon: DIR *%c* TYPE *%c* NAME *%-15s* ",
                                          lc->DIRECTION,lc->TYPE,lc->NAME );
         (void)fprintf(stdout,"sig: TYPE *%c* INDEX *%6ld* NAME *%-15s* \n",
                                          lc->SIG->TYPE,lc->SIG->INDEX,getsigname(lc->SIG) ) ;
#     endif

      lc->SIG->TYPE = EXTERNAL; /* probably not necessary */

      onexternalbuf = 0;
      for (c = (chain_list *) ( getptype( lc->SIG->USER, (long) LOFIGCHAIN )->DATA ) ;
           c;
           c = c->NEXT) {
#     ifdef DEBUG
           (void)fprintf(stdout,"chain: NAME *%s* ", ( (locon_list *) c->DATA)->NAME );
           (void)fprintf(stdout,"DIR *%c* TYPE *%c*",
                   ( (locon_list *)c->DATA )->DIRECTION, ( (locon_list *)c->DATA )->TYPE );
           if (  (((locon_list *)c->DATA )->TYPE ) == INTERNAL ) {
              (void)fprintf(stdout,"FIGNAME %s\n",
                              ( (loins_list *) ((locon_list *)c->DATA )->ROOT )->FIGNAME );
           } else {
              (void)fprintf(stdout,"****\n");
           }
#     endif
         if (((locon_list *)c->DATA)->TYPE == EXTERNAL) {
            externalcon = (locon_list *)c->DATA;
            continue;
	  }
         if (((locon_list *)c->DATA)->TYPE != INTERNAL)
            continue;
         li = (loins_list *)((locon_list *)c->DATA)->ROOT;
         for( i=0 ; i < spec_ibuf_size ; i++ ) {
            if ( ! strcmp(li->FIGNAME, spec_ibuf[i] ) ) {
               onexternalbuf++;
               break ;
            }
         }
       }
       if (onexternalbuf) {
         externalcon->USER = (ptype_list *)-1;  /* ugly (but fast) */
       }
   }

   for (index = 0, ls = lof->LOSIG; ls; ls = ls->NEXT) {

#     ifdef DEBUG
         (void)fprintf(stdout, "losig : TYPE *%c* index *%6ld* name *%s*\n",
                                          ls->TYPE, ls->INDEX, getsigname(ls) );
#     endif
      if (index < ls->INDEX)
         index = ls->INDEX;
      if (getsigname(ls) == VDD)
         sigvdd = ls;
      if (getsigname(ls) == VSS)
         sigvss = ls;
      hsize++;
 
   }
   index++;
   if (!sigvdd || !sigvss) {
      fflush(stdout);
      fprintf(stderr, "mbk2xnf error: no vdd or vss signals!\n");
      EXIT(1);
   }

   h = addht(hsize << 1); /* bigger means faster! */

   /* add IO buffers and rename the tristate ones connected to output
      ports. */
   for (lc = lof->LOCON; lc; lc = lc->NEXT) {
   chain_list *c;

      if (isvdd(lc->NAME) || isvss(lc->NAME))
         continue;
      c = addchain(NULL, (void *)sigvss);
      c = addchain(c, (void *)sigvdd);
      switch(lc->DIRECTION) {
         case IN:

            if (lc->USER == (ptype_list *)-1) {
               freechain(c);
               continue;
	     }

            c = addchain(c, (void *)lc->SIG);
            lc->SIG->TYPE = INTERNAL;
            sprintf(buffer1, "$%s", lc->NAME);
            lc->SIG = addlosig(lof, index++,
                              addchain(NULL, (void *)buffer1),
                              EXTERNAL, 0.0);
            c = addchain(c, (void *)lc->SIG);
            addloins(lof, buffer1, getlomodel(buflist, ibuf), c); 
            break;
            
         case INOUT:
            lc->DIRECTION = OUT;
         case OUT:
            ls = lc->SIG;
            lc->SIG->TYPE = INTERNAL;
            sprintf(buffer1, "$%s", lc->NAME);
            lc->SIG = addlosig(lof, index++,
                              addchain(NULL, (void *)buffer1),
                              EXTERNAL, 0.0);
            c = addchain(c, (void *)lc->SIG);
            c = addchain(c, (void *)ls);
            addloins(lof, buffer1, getlomodel(buflist, obuf), c); 
            break;

         case TRISTATE:
            for (c = (chain_list *)getptype(lc->SIG->USER, LOFIGCHAIN)->DATA;
                 c;
                 c = c->NEXT) {
               if (((locon_list *)c->DATA)->TYPE != INTERNAL)
                  continue;
               if (((loins_list *)((locon_list *)c->DATA)->ROOT)->FIGNAME == tbuf)
                  ((loins_list *)((locon_list *)c->DATA)->ROOT)->FIGNAME = obuft;
            }
            break;

         default:
            fflush(stdout);
            fprintf(stderr, "mbk2xnf error: unsupported external ");
            fprintf(stderr, "port direction on %s\n", lc->NAME);
            EXIT(1);
      }
   }

   for (li = lof->LOINS; li; li = li->NEXT) {
      fprintf(file, "SYM,%s,%s,\n", xnfvector( li->INSNAME, buffer1 ),
                                    xnf_map_figname( li->FIGNAME, buffer2 ) );
      for(lc = li->LOCON; lc; lc = lc->NEXT) {
         /* no vdd/vss required in this world, as far as I can tell */
         if (!isvdd(lc->NAME) && !isvss(lc->NAME)) {
            fprintf(file, "PIN,%s,%c,%s,", prim_map_pin_name( lc->NAME, buffer1 ), dir(lc),
                                           signalname( h, lof, lc->SIG, buffer2 )    );
            if ( prim_add_inv(lc->NAME) ) {
               fprintf(file,",INV\n");
            } else {
              fprintf(file,"\n"); 
            }
	 }
      }
      fprintf(file, "END\n");
   }
   /* since we've played around with the external connectors, we must
      drive the correct external signal names... */
   for (lc = lof->LOCON; lc; lc = lc->NEXT)
      if (!isvdd(getsigname(lc->SIG)) && !isvss(getsigname(lc->SIG)))
         fprintf(file, "EXT,%s,%c,\n", signalname( h, lof, lc->SIG, buffer1 ), dir(lc));

   fprintf(file, "EOF\n");
   delht(h);
   fclose(file);
}
